@@ -24,7 +24,18 @@
 #include <map>
 #include <list>
 
-#define classad_hash_map std::map 
-#define classad_slist    std::list
+#ifdef WIN32
+#include <hash_map>
+#else
+#include <ext/hash_map>
+#endif
+
+#ifdef WIN32
+#define classad_hash_map stdext::hash_map
+#else
+#define classad_hash_map __gnu_cxx::hash_map
+#endif
+#define classad_map   std::map 
+#define classad_slist std::list
 
 #endif /* __CLASSAD_STL_H__ */
