@@ -28,6 +28,11 @@
 #include "classad/sink.h"
 #include "classad/util.h"
 
+#if defined(WIN32) && !defined(USE_PCRE) && !defined(USE_POSIX_REGEX)
+  #define USE_PCRE
+  #define HAVE_PCRE_H
+#endif
+
 #if defined USE_POSIX_REGEX 
   #include <regex.h>
 #elif defined USE_PCRE
@@ -156,6 +161,7 @@ FunctionCall( )
 			// for compatibility with old classads:
 		functionTable["ifThenElse"  ] = (void*)ifThenElse;
 		functionTable["interval" ] = (void*)interval;
+		functionTable["eval"] = (void*)eval;
 
 		initialized = true;
 	}
@@ -228,14 +234,14 @@ bool FunctionCall::
 SameAs(const ExprTree *tree) const
 {
     bool is_same;
-    FunctionCall *other_fn;
+    const FunctionCall *other_fn;
     
     if (this == tree) {
         is_same = true;
     } else if (tree->GetKind() != FN_CALL_NODE) {
         is_same = false;
     } else {
-        other_fn = (FunctionCall *) tree;
+        other_fn = (const FunctionCall *) tree;
         
         if (functionName == other_fn->functionName
             && function == other_fn->function
@@ -476,6 +482,7 @@ _Flatten( EvalState &state, Value &value, ExprTree*&tree, int* ) const
 		return false;
 	}
 	newCall->functionName = functionName;
+	newCall->function = function;
 
 	// flatten the arguments
 	for(ArgumentList::const_iterator i=arguments.begin();i!=arguments.end();
@@ -1161,18 +1168,7 @@ timeZoneOffset (const char *, const ArgumentList &argList, EvalState &,
 		return( true );
 	}
 	
-	struct tm *tms;
-	time_t clock;
-	time(&clock);
-	tms  = localtime( &clock);
-		// POSIX says that timezone is positive west of GMT;  we reverse it
-		// here to make it more intuitive
-	if(tms->tm_isdst > 0) { //check for daylight saving
-	  val.SetRelativeTimeValue( (time_t) (-timezone_offset()+3600) );
-	}
-	else {
-	  val.SetRelativeTimeValue( (time_t) (-timezone_offset()));
-	}
+	val.SetRelativeTimeValue( (time_t) timezone_offset( time(NULL), false ) );
 	return( true );
 }
 
@@ -1206,7 +1202,7 @@ makeDate( const char*, const ArgumentList &argList, EvalState &state,
 	Value 	arg0, arg1, arg2;
 	int		dd, mm, yy;
 	time_t	clock;
-	struct	tm	*tms;
+	struct	tm	tms;
 	char	buffer[64];
 	string	month;
 	abstime_t abst;
@@ -1229,7 +1225,7 @@ makeDate( const char*, const ArgumentList &argList, EvalState &state,
 		val.SetErrorValue( );
 		return( false );
 	}
-	tms  = localtime( &clock);
+	getLocalTime( &clock, &tms );
 
 		// evaluate optional third argument
 	if( argList.size( ) == 3 ) {
@@ -1239,7 +1235,7 @@ makeDate( const char*, const ArgumentList &argList, EvalState &state,
 		}
 	} else {
 			// use the current year (tm_year is years since 1900)
-		arg2.SetIntegerValue( tms->tm_year + 1900 );
+		arg2.SetIntegerValue( tms.tm_year + 1900 );
 	}
 		
 		// check if any of the arguments are undefined
@@ -1265,7 +1261,7 @@ makeDate( const char*, const ArgumentList &argList, EvalState &state,
 		}
 	} else if( arg1.IsStringValue( month ) ) {
 		if( sprintf( buffer, "%d %s %d", dd, month.c_str( ), yy ) > 63 ||
-				strptime( buffer, "%d %b %Y", tms ) == NULL ) {
+				strptime( buffer, "%d %b %Y", &tms ) == NULL ) {
 			val.SetErrorValue( );
 			return( true );
 		}
@@ -1275,13 +1271,13 @@ makeDate( const char*, const ArgumentList &argList, EvalState &state,
 	}
 
 		// convert the struct tm -> time_t -> absolute time
-	clock = mktime( tms );
+	clock = mktime( &tms );
 	if(clock == -1) {
 		val.SetErrorValue( );
 		return( true );
 	}
 	abst.secs = clock;	
-	abst.offset = -timezone_offset();
+	abst.offset = timezone_offset();
 	// alter absolute time parameters based on current day-light saving status
 	if(tms->tm_isdst > 0) { 
 		abst.offset += 3600;
@@ -1490,6 +1486,8 @@ formatTime(const char*, const ArgumentList &argList, EvalState &state,
     return did_eval;
 }
 
+#if 0
+
 bool FunctionCall::
 inTimeUnits(const char*name,const ArgumentList &argList,EvalState &state, 
 	Value &val )
@@ -1541,6 +1539,8 @@ inTimeUnits(const char*name,const ArgumentList &argList,EvalState &state,
 	return( true );
 }
 
+#endif
+
 	// concatenate all arguments (expected to be strings)
 bool FunctionCall::
 strCat( const char*, const ArgumentList &argList, EvalState &state, 
@@ -1627,10 +1627,10 @@ changeCase(const char*name,const ArgumentList &argList,EvalState &state,
             return true;
         } else if (stringVal.IsErrorValue()) {
             result.SetErrorValue();
-            return false;
+            return true;
         } else if (!stringVal.IsStringValue(str)) {
             result.SetErrorValue();
-            return false;
+            return true;
         }
 	}
 
@@ -1654,7 +1654,7 @@ subString( const char*, const ArgumentList &argList, EvalState &state,
 		// two or three arguments
 	if( argList.size() < 2 || argList.size() > 3 ) {
 		result.SetErrorValue( );
-		return( false );
+		return( true );
 	}
 
 		// Evaluate all arguments
@@ -1669,14 +1669,14 @@ subString( const char*, const ArgumentList &argList, EvalState &state,
 	if( arg0.IsUndefinedValue( ) || arg1.IsUndefinedValue( ) ||
 		(argList.size() > 2 && arg2.IsUndefinedValue( ) ) ) {
 		result.SetUndefinedValue( );
-		return( false );
+		return( true );
 	}
 
 		// arg0 must be string, arg1 must be int, arg2 (if given) must be int
 	if( !arg0.IsStringValue( buf ) || !arg1.IsIntegerValue( offset )||
 		(argList.size( ) > 2 && !arg2.IsIntegerValue( len ) ) ) {
 		result.SetErrorValue( );
-		return( false );
+		return( true );
 	}
 
 		// perl-like substr; negative offsets and lengths count from the end
@@ -1689,6 +1689,9 @@ subString( const char*, const ArgumentList &argList, EvalState &state,
 	}
 	if( len <= 0 ) {
 		len = alen - offset + len;
+		if( len < 0 ) {
+			len = 0;
+		}
 	} else if( len > alen - offset ) {
 		len = alen - offset;
 	}
@@ -1720,7 +1723,7 @@ compareString( const char*name, const ArgumentList &argList, EvalState &state,
     // Must have two arguments
 	if(argList.size() != 2) {
 		result.SetErrorValue( );
-		return( false );
+		return( true );
 	}
 
     // Evaluate both arguments
@@ -1734,7 +1737,7 @@ compareString( const char*name, const ArgumentList &argList, EvalState &state,
     // undefined.
 	if(arg0.IsUndefinedValue() || arg1.IsUndefinedValue()) {
 		result.SetUndefinedValue( );
-		return false;
+		return true;
     }
 
     string  s0, s1;
@@ -1932,7 +1935,7 @@ convTime(const char* name,const ArgumentList &argList,EvalState &state,
 		time_t rsecs = 0;
 		if(relative) {// 2nd argument is N/A for reltime
 			result.SetErrorValue( );
-			return( false );
+			return( true );
 		}
 		// 2nd arg should be integer, real or reltime
 		else if (arg2.IsIntegerValue(ivalue2)) {
@@ -1946,7 +1949,7 @@ convTime(const char* name,const ArgumentList &argList,EvalState &state,
 		}
 		else {
 			result.SetErrorValue( );
-			return( false );
+			return( true );
 		}
 	} else {
         secondarg = false;
@@ -2112,11 +2115,12 @@ random( const char*,const ArgumentList &argList,EvalState &state,
     double  random_double;
 
     // takes exactly one argument
-	if( argList.size() != 1 ) {
+	if( argList.size() > 1 ) {
 		result.SetErrorValue( );
 		return( true );
-	}
-	if( !argList[0]->Evaluate( state, arg ) ) {
+	} else if ( argList.size() == 0 ) {
+		arg.SetRealValue( 1.0 );
+	} else if( !argList[0]->Evaluate( state, arg ) ) {
 		result.SetErrorValue( );
 		return( false );
 	}
@@ -2135,7 +2139,7 @@ random( const char*,const ArgumentList &argList,EvalState &state,
 }
 
 bool FunctionCall::
-ifThenElse( const char* name,const ArgumentList &argList,EvalState &state,
+ifThenElse( const char* /* name */,const ArgumentList &argList,EvalState &state,
 	Value &result )
 {
 	Value	arg1;
@@ -2168,8 +2172,8 @@ ifThenElse( const char* name,const ArgumentList &argList,EvalState &state,
 		break;
 	}
 	case Value::REAL_VALUE: {
-		int realval;
-		if( !arg1.IsIntegerValue(realval) ) {
+		double realval;
+		if( !arg1.IsRealValue(realval) ) {
 			result.SetErrorValue();
 			return( false );
 		}
@@ -2209,7 +2213,47 @@ ifThenElse( const char* name,const ArgumentList &argList,EvalState &state,
 }
 
 bool FunctionCall::
-interval( const char* name,const ArgumentList &argList,EvalState &state,
+eval( const char* /* name */,const ArgumentList &argList,EvalState &state,
+	  Value &result )
+{
+	Value arg,strarg;
+
+		// takes exactly one argument
+	if( argList.size() != 1 ) {
+		result.SetErrorValue( );
+		return true;
+	}
+	if( !argList[0]->Evaluate( state, arg ) ) {
+		result.SetErrorValue( );
+		return false;
+	}
+
+	string s;
+    if( !convertValueToStringValue(arg, strarg) ||
+		!strarg.IsStringValue( s ) )
+	{
+		result.SetErrorValue();
+		return true;
+	}
+
+	ClassAdParser parser;
+	ExprTree *expr = NULL;
+	if( !parser.ParseExpression( s.c_str(), expr, true ) || !expr ) {
+		result.SetErrorValue();
+		return true;
+	}
+
+	expr->SetParentScope( state.curAd );
+
+	if( !expr->Evaluate( result ) ) {
+		result.SetErrorValue();
+		return false;
+	}
+	return true;
+}
+
+bool FunctionCall::
+interval( const char* /* name */,const ArgumentList &argList,EvalState &state,
 	Value &result )
 {
 	Value	arg,intarg;
@@ -2697,11 +2741,11 @@ doSplitTime(const Value &time, ClassAd * &splitClassAd)
     did_conversion = true;
     if (time.IsIntegerValue(integer)) {
         asecs.secs = integer;
-        asecs.offset = timezone_offset();
+        asecs.offset = timezone_offset( asecs.secs, false );
         absTimeToClassAd(asecs, splitClassAd);
     } else if (time.IsRealValue(real)) {
         asecs.secs = (int) real;
-        asecs.offset = timezone_offset();
+        asecs.offset = timezone_offset( asecs.secs, false );
         absTimeToClassAd(asecs, splitClassAd);
     } else if (time.IsAbsoluteTimeValue(asecs)) {
         absTimeToClassAd(asecs, splitClassAd);
@@ -2724,7 +2768,7 @@ absTimeToClassAd(const abstime_t &asecs, ClassAd * &splitClassAd)
 
     splitClassAd = new ClassAd;
 
-    clock = asecs.secs;
+    clock = asecs.secs + asecs.offset;
     getGMTime( &clock, &tms );
 
     splitClassAd->InsertAttr("Type", "AbsoluteTime");
