@@ -29,6 +29,11 @@ extern "C" void to_lower (char *);	// from util_lib (config.c)
 
 BEGIN_NAMESPACE( classad )
 
+// This flag is only meant for use in Condor, which is transitioning
+// from an older version of ClassAds with slightly different evaluation
+// semantics. It will be removed without warning in a future release.
+bool _useOldClassAdSemantics = false;
+
 // This is probably not the best place to put these. However, 
 // I am reconsidering how we want to do errors, and this may all
 // change in any case. 
@@ -39,13 +44,13 @@ void ClassAdLibraryVersion(int &major, int &minor, int &patch)
 {
     major = 1;
     minor = 0;
-    patch = 1;
+    patch = 8;
     return;
 }
 
 void ClassAdLibraryVersion(string &version_string)
 {
-    version_string = "1.0.3";
+    version_string = "1.0.8";
     return;
 }
 
@@ -55,6 +60,7 @@ ClassAd ()
 	nodeKind = CLASSAD_NODE;
 	EnableDirtyTracking();
 	chained_parent_ad = NULL;
+	alternateScope = NULL;
 }
 
 
@@ -89,6 +95,7 @@ CopyFrom( const ClassAd &ad )
 		Clear( );
         ExprTree::CopyFrom(ad);
 		chained_parent_ad = ad.chained_parent_ad;
+		alternateScope = ad.alternateScope;
 		
 		DisableDirtyTracking();
 		for( itr = ad.attrList.begin( ); itr != ad.attrList.end( ); itr++ ) {
@@ -139,9 +146,9 @@ SameAs(const ExprTree *tree) const
    } else if (tree->GetKind() != CLASSAD_NODE) {
         is_same = false;
    } else {
-       ClassAd *other_classad;
+       const ClassAd *other_classad;
 
-       other_classad = (ClassAd *) tree;
+       other_classad = (const ClassAd *) tree;
 
        if (attrList.size() != other_classad->attrList.size()) {
            is_same = false;
@@ -353,13 +360,14 @@ Insert( const string &name, ExprTree *tree )
 	// parent of the expression is this classad
 	tree->SetParentScope( this );
 
-		// check if attribute already exists in classad
-	AttrList::iterator itr = attrList.find( name );
-	if( itr != attrList.end( ) ) {
-			// delete old expression and replace
-		delete itr->second;
+	pair<AttrList::iterator,bool> insert_result =
+		attrList.insert( AttrList::value_type(name,tree) );
+
+	if( !insert_result.second ) {
+			// replace existing value
+		delete insert_result.first->second;
+		insert_result.first->second = tree;
 	}
-	attrList[name] = tree;
 
 	MarkAttributeDirty(name);
         
@@ -432,7 +440,7 @@ int ClassAd::
 LookupInScope(const string &name, ExprTree*& expr, EvalState &state) const
 {
 	extern int exprHash( const ExprTree* const&, int );
-	ClassAd 		*current = (ClassAd*)this, *superScope;
+	const ClassAd *current = this, *superScope;
 	Value			val;
 
 	expr = NULL;
@@ -447,18 +455,18 @@ LookupInScope(const string &name, ExprTree*& expr, EvalState &state) const
 			return( EVAL_OK );
 		}
 
-		superScope = (ClassAd*)current->parentScope;
+		superScope = current->parentScope;
 		if(strcasecmp(name.c_str( ),"toplevel")==0 || 
 				strcasecmp(name.c_str( ),"root")==0){
 			// if the "toplevel" attribute was requested ...
-			expr = state.rootAd;
+			expr = (ClassAd*)state.rootAd;
 			if( expr == NULL ) {	// NAC - circularity so no root
 				return EVAL_FAIL;  	// NAC
 			}						// NAC
 			return( expr ? EVAL_OK : EVAL_UNDEF );
 		} else if( strcasecmp( name.c_str( ), "self" ) == 0 ) {
 			// if the "self" ad was requested
-			expr = state.curAd;
+			expr = (ClassAd*)state.curAd;
 			return( expr ? EVAL_OK : EVAL_UNDEF );
 		} else if( strcasecmp( name.c_str( ), "parent" ) == 0 ) {
 			// the lexical parent
@@ -722,13 +730,13 @@ _Flatten( EvalState& state, Value&, ExprTree*& tree, int* ) const
 	ClassAd 	*newAd = new ClassAd();
 	Value		eval;
 	ExprTree	*etree;
-	ClassAd		*oldAd;
+	const ClassAd *oldAd;
 	AttrList::const_iterator	itr;
 
 	tree = NULL; // Just to be safe...  wenger 2003-12-11.
 
 	oldAd = state.curAd;
-	state.curAd = (ClassAd*)this;
+	state.curAd = this;
 
 	for( itr = attrList.begin( ); itr != attrList.end( ); itr++ ) {
 		// flatten expression
@@ -912,7 +920,7 @@ GetExternalReferences( const ExprTree *tree, References &refs, bool fullNames )
     EvalState       state;
 
     state.rootAd = this; 
-    state.curAd = (ClassAd*)tree->GetParentScope( );
+    state.curAd = tree->GetParentScope( );
 	if( !state.curAd ) state.curAd = this;
 	
     return( _GetExternalReferences( tree, this, state, refs, fullNames ) );
@@ -929,13 +937,13 @@ _GetExternalReferences( const ExprTree *expr, ClassAd *ad,
             return( true );
 
         case ATTRREF_NODE: {
-            ClassAd   		*start;
+            const ClassAd   *start;
             ExprTree        *tree, *result;
             string          attr;
             Value           val;
             bool            abs;
 
-            ((AttributeReference*)expr)->GetComponents( tree, attr, abs );
+            ((const AttributeReference*)expr)->GetComponents( tree, attr, abs );
                 // establish starting point for attribute search
             if( tree==NULL ) {
                 start = abs ? state.rootAd : state.curAd;
@@ -971,7 +979,7 @@ _GetExternalReferences( const ExprTree *expr, ClassAd *ad,
                 }
             }
                 // lookup for attribute
-			ClassAd *curAd = state.curAd;
+			const ClassAd *curAd = state.curAd;
             switch( start->LookupInScope( attr, result, state ) ) {
                 case EVAL_ERROR:
                         // some error
@@ -1000,7 +1008,7 @@ _GetExternalReferences( const ExprTree *expr, ClassAd *ad,
                 // recurse on subtrees
             Operation::OpKind	op;
             ExprTree    		*t1, *t2, *t3;
-            ((Operation*)expr)->GetComponents( op, t1, t2, t3 );
+            ((const Operation*)expr)->GetComponents( op, t1, t2, t3 );
             if( t1 && !_GetExternalReferences( t1, ad, state, refs, fullNames ) ) {
                 return( false );
             }
@@ -1019,7 +1027,7 @@ _GetExternalReferences( const ExprTree *expr, ClassAd *ad,
             vector<ExprTree*>           args;
             vector<ExprTree*>::iterator itr;
 
-            ((FunctionCall*)expr)->GetComponents( fnName, args );
+            ((const FunctionCall*)expr)->GetComponents( fnName, args );
             for( itr = args.begin( ); itr != args.end( ); itr++ ) {
                 if( !_GetExternalReferences( *itr, ad, state, refs, fullNames ) ) {
 					return( false );
@@ -1034,7 +1042,7 @@ _GetExternalReferences( const ExprTree *expr, ClassAd *ad,
             vector< pair<string, ExprTree*> >           attrs;
             vector< pair<string, ExprTree*> >::iterator itr;
 
-            ((ClassAd*)expr)->GetComponents( attrs );
+            ((const ClassAd*)expr)->GetComponents( attrs );
             for( itr = attrs.begin( ); itr != attrs.end( ); itr++ ) {
                 if( !_GetExternalReferences( itr->second, ad, state, refs, fullNames )) {
 					return( false );
@@ -1049,7 +1057,7 @@ _GetExternalReferences( const ExprTree *expr, ClassAd *ad,
             vector<ExprTree*>           exprs;
             vector<ExprTree*>::iterator itr;
 
-            ((ExprList*)expr)->GetComponents( exprs );
+            ((const ExprList*)expr)->GetComponents( exprs );
             for( itr = exprs.begin( ); itr != exprs.end( ); itr++ ) {
                 if( !_GetExternalReferences( *itr, ad, state, refs, fullNames ) ) {
 					return( false );
@@ -1070,7 +1078,7 @@ GetExternalReferences( const ExprTree *tree, PortReferences &refs )
     EvalState       state;
 
     state.rootAd = this; 
-    state.curAd = (ClassAd*)tree->GetParentScope( );
+    state.curAd = tree->GetParentScope( );
 	if( !state.curAd ) state.curAd = this;
 	
     return( _GetExternalReferences( tree, this, state, refs ) );
@@ -1086,14 +1094,14 @@ _GetExternalReferences( const ExprTree *expr, ClassAd *ad,
             return( true );
 
         case ATTRREF_NODE: {
-            ClassAd   		*start;
+            const ClassAd   *start;
             ExprTree        *tree, *result;
             string          attr;
             Value           val;
             bool            abs;
 			PortReferences::iterator	pitr;
 
-            ((AttributeReference*)expr)->GetComponents( tree, attr, abs );
+            ((const AttributeReference*)expr)->GetComponents( tree, attr, abs );
                 // establish starting point for attribute search
             if( tree==NULL ) {
                 start = abs ? state.rootAd : state.curAd;
@@ -1119,7 +1127,7 @@ _GetExternalReferences( const ExprTree *expr, ClassAd *ad,
 				}
             }
                 // lookup for attribute
-			ClassAd *curAd = state.curAd;
+			const ClassAd *curAd = state.curAd;
             switch( start->LookupInScope( attr, result, state ) ) {
                 case EVAL_ERROR:
                         // some error
@@ -1148,7 +1156,7 @@ _GetExternalReferences( const ExprTree *expr, ClassAd *ad,
                 // recurse on subtrees
             Operation::OpKind   op;
             ExprTree    		*t1, *t2, *t3;
-            ((Operation*)expr)->GetComponents( op, t1, t2, t3 );
+            ((const Operation*)expr)->GetComponents( op, t1, t2, t3 );
             if( t1 && !_GetExternalReferences( t1, ad, state, refs ) ) {
                 return( false );
             }
@@ -1167,7 +1175,7 @@ _GetExternalReferences( const ExprTree *expr, ClassAd *ad,
             vector<ExprTree*>           args;
             vector<ExprTree*>::iterator itr;
 
-            ((FunctionCall*)expr)->GetComponents( fnName, args );
+            ((const FunctionCall*)expr)->GetComponents( fnName, args );
             for( itr = args.begin( ); itr != args.end( ); itr++ ) {
                 if( !_GetExternalReferences( *itr, ad, state, refs ) ) {
 					return( false );
@@ -1182,7 +1190,7 @@ _GetExternalReferences( const ExprTree *expr, ClassAd *ad,
             vector< pair<string, ExprTree*> >           attrs;
             vector< pair<string, ExprTree*> >::iterator itr;
 
-            ((ClassAd*)expr)->GetComponents( attrs );
+            ((const ClassAd*)expr)->GetComponents( attrs );
             for( itr = attrs.begin( ); itr != attrs.end( ); itr++ ) {
                 if( !_GetExternalReferences( itr->second, ad, state, refs )) {
 					return( false );
@@ -1197,7 +1205,7 @@ _GetExternalReferences( const ExprTree *expr, ClassAd *ad,
             vector<ExprTree*>           exprs;
             vector<ExprTree*>::iterator itr;
 
-            ((ExprList*)expr)->GetComponents( exprs );
+            ((const ExprList*)expr)->GetComponents( exprs );
             for( itr = exprs.begin( ); itr != exprs.end( ); itr++ ) {
                 if( !_GetExternalReferences( *itr, ad, state, refs ) ) {
 					return( false );
@@ -1212,6 +1220,224 @@ _GetExternalReferences( const ExprTree *expr, ClassAd *ad,
     }
 }
 
+
+bool ClassAd::
+GetInternalReferences( const ExprTree *tree, References &refs, bool fullNames)
+{
+    EvalState state;
+    state.rootAd = this;
+    state.curAd = tree->GetParentScope();
+    if(!state.curAd) state.curAd = this;
+
+    return( _GetInternalReferences( tree, this, state, refs, fullNames) );
+}
+
+//this is closely modelled off of _GetExternalReferences in the new_classads.
+bool ClassAd::
+_GetInternalReferences( const ExprTree *expr, ClassAd *ad,
+        EvalState &state, References& refs, bool fullNames)
+{
+
+    switch( expr->GetKind() ){
+        //nothing to be found here!
+        case LITERAL_NODE:{
+            return true;
+        break;
+                          }
+
+        case ATTRREF_NODE:{
+            const ClassAd   *start;
+            ExprTree        *tree, *result;
+            string          attr;
+            Value           val;
+            bool            abs;
+
+            ((const AttributeReference*)expr)->GetComponents(tree,attr,abs);
+
+            //figuring out which state to base this off of
+            if( tree == NULL ){
+                start = abs ? state.rootAd : state.curAd;
+                //remove circularity
+                if( abs && (start == NULL) ) {
+                    return false;
+                }
+            } else {
+                if (!tree->Evaluate(state, val) ) {
+                    return false;
+                }
+
+                /*
+                 *
+                 * [
+                 * A = 3;
+                 * B = { 1, 3, 5};
+                 * C = D.A + 1;
+                 * D = [ A = E; B = 4; ];
+                 * E = 4;
+                 * ];
+                 */
+
+                if( val.IsUndefinedValue() ) {
+                    return true;
+                }
+
+                string nameToAddToRefs = "";
+
+                string prefixStr;
+
+                if(tree != NULL)
+                {
+                    ClassAdUnParser unparser;
+                    //TODO: figure out how this would handle "B.D.G"
+                    unparser.Unparse(prefixStr, tree);
+                    //fullNameCpy = prefixStr;
+                    nameToAddToRefs = prefixStr;
+
+                    if(fullNames)
+                    {
+                        nameToAddToRefs += ".";
+                        nameToAddToRefs += attr;
+                    }
+                }
+
+                //WOULD INSERT "" IF tree == NULL! BAD! FIXME
+                refs.insert(nameToAddToRefs);
+
+                ExprTree *followRef;
+                //TODO: If we get to this point, must there be a prefix?
+                //  FIGURE OUT WHAT A SIMPLE / ABSOLUTE ATTR IS
+                followRef = ad->Lookup(prefixStr);
+                //this is checking to see if the prefix is in
+                //  this classad or not.
+                // if not, then we just let the loop continue
+                if(followRef != NULL)
+                {
+                    return _GetInternalReferences(followRef, ad,
+                                        state, refs, fullNames);
+                }
+
+                //otherwise, if the tree didn't evaluate to a classad,
+                //we have a problemo, mon.
+                //TODO: but why?
+                if( !val.IsClassAdValue( start ) ) {
+                    return false;
+                }
+            }
+
+            const ClassAd *curAd = state.curAd;
+            switch( start->LookupInScope( attr, result, state) ) {
+                case EVAL_ERROR:
+                    return false;
+                break;
+
+                //attr is external, so let's find the internals in that
+                //result
+                //JUST KIDDING
+                case EVAL_UNDEF:{
+                    
+                    //bool rval = _GetInternalReferences(result, ad, state, refs, fullNames);
+                    //state.curAd = curAd;
+                    return true;
+                break;
+                                }
+
+                case EVAL_OK:   {
+                    //whoo, it's internal.
+                    refs.insert(attr);
+                    bool rval =_GetInternalReferences(result, ad, state, refs, fullNames);
+                    //TODO: Does this actually matter?
+                    state.curAd = curAd;
+                    return rval;
+                break;
+                                }
+
+                default:
+                    // "enh??"
+                    return false;
+                break;
+            }
+
+        break;
+                          }
+    
+
+        case OP_NODE:{
+
+            //recurse on subtrees
+            Operation::OpKind       op;
+            ExprTree                *t1, *t2, *t3;
+            ((const Operation*)expr)->GetComponents(op, t1, t2, t3);
+            if( t1 && !_GetInternalReferences(t1, ad, state, refs, fullNames)) {
+                return false;
+            }
+
+            if( t2 && !_GetInternalReferences(t2, ad, state, refs, fullNames)) {
+                return false;
+            }
+
+            if( t3 && !_GetInternalReferences(t3, ad, state, refs, fullNames)) {
+                return false;
+            }
+
+            return true;
+        break;
+        }
+
+        case FN_CALL_NODE:{
+            //recurse on the subtrees!
+            string                          fnName;
+            vector<ExprTree*>               args;
+            vector<ExprTree*>::iterator     itr;
+
+            ((const FunctionCall*)expr)->GetComponents(fnName, args);
+            for( itr = args.begin(); itr != args.end(); itr++){
+                if( !_GetInternalReferences( *itr, ad, state, refs, fullNames) ) {
+                    return false;
+                }
+            }
+
+            return true;
+        break;
+                          }
+
+        case CLASSAD_NODE:{
+            //also recurse on subtrees...
+            vector< pair<string, ExprTree*> >               attrs;
+            vector< pair<string, ExprTree*> >:: iterator    itr;
+
+            ((const ClassAd*)expr)->GetComponents(attrs);
+            for(itr = attrs.begin(); itr != attrs.end(); itr++){
+                if( !_GetInternalReferences(itr->second, ad, state, refs, fullNames)) {
+                    return false;
+                }
+            }
+
+            return true;
+        break;
+            }
+
+        case EXPR_LIST_NODE:{
+            vector<ExprTree*>               exprs;
+            vector<ExprTree*>::iterator     itr;
+
+            ((const ExprList*)expr)->GetComponents(exprs);
+            for(itr = exprs.begin(); itr != exprs.end(); itr++){
+                if( !_GetInternalReferences(*itr, ad, state, refs, fullNames) ) {
+                    return false;
+                }
+            }
+
+            return true;
+        break;
+                            }
+
+        default:
+            return false;
+
+    }
+
+}
+
 #if defined( EXPERIMENTAL )
 
 bool ClassAd::
