@@ -104,12 +104,12 @@ bool Operation::
 SameAs(const ExprTree *tree) const
 {
     bool is_same;
-    Operation *other_op;
+    const Operation *other_op;
 
     if (tree->GetKind() != OP_NODE) {
         is_same = false;
     } else {
-        other_op = (Operation *) tree;
+        other_op = (const Operation *) tree;
         
         if (   operation == other_op->operation
             && SameChild(child1, other_op->child1)
@@ -259,13 +259,10 @@ _doOperation (OpKind op, Value &val1, Value &val2, Value &val3,
 			return SIG_CHLD1;
 		}
 
-		// otherwise, if the selector is not boolean propagate error
-		if( vt1 != Value::BOOLEAN_VALUE ) {
+		if( !val1.IsBooleanValueEquiv(b) ) {
 			result.SetErrorValue();	
 			return SIG_CHLD1;
-		}
-
-		if( val1.IsBooleanValue(b) && b ) {
+		} else if( b ) {
 			result.CopyFrom( val2 );
 			return( SIG_CHLD2 );
 		} else {
@@ -322,7 +319,6 @@ _doOperation (OpKind op, Value &val1, Value &val2, Value &val3,
 	return SIG_NONE;
 }
 
-
 bool Operation::
 _Evaluate (EvalState &state, Value &result) const
 {
@@ -341,6 +337,10 @@ _Evaluate (EvalState &state, Value &result) const
 			return( false );
 		}
 		valid1 = true;
+
+		if( shortCircuit( state, val1, result ) ) {
+			return true;
+		}
 	}
 
 	if (child2) {
@@ -364,6 +364,44 @@ _Evaluate (EvalState &state, Value &result) const
 	return( rval != SIG_NONE );
 }
 
+bool Operation::
+shortCircuit( EvalState &state, Value const &arg1, Value &result ) const
+{
+	bool arg1_bool;
+
+	switch( operation ) {
+	case LOGICAL_OR_OP:
+		if( arg1.IsBooleanValueEquiv(arg1_bool) && arg1_bool ) {
+			result.SetBooleanValue( true );
+			return true;
+		}
+		break;
+	case LOGICAL_AND_OP:
+		if( arg1.IsBooleanValueEquiv(arg1_bool) && !arg1_bool ) {
+			result.SetBooleanValue( false );
+			return true;
+		}
+		break;
+	case TERNARY_OP:
+		if( arg1.IsBooleanValueEquiv(arg1_bool) ) {
+			if( arg1_bool ) {
+				if( child2 ) {
+					return child2->Evaluate(state,result);
+				}
+			}
+			else {
+				if( child3 ) {
+					return child3->Evaluate(state,result);
+				}
+			}
+		}
+		break;
+	default:
+		// no-op
+		break;
+	}
+	return false;
+}
 
 bool Operation::
 _Evaluate( EvalState &state, Value &result, ExprTree *& tree ) const
@@ -551,10 +589,14 @@ _Flatten( EvalState &state, Value &val, ExprTree *&tree, int *opPtr ) const
 	// Flatten sub expressions
 	if( child1 && !child1->Flatten( state, val1, fChild1, &childOp1 ) ||
 		child2 && !child2->Flatten( state, val2, fChild2, &childOp2 ) ) {
+		delete fChild1;
+		delete fChild2;
 		tree = NULL;
 		return false;
 	}
 
+		// NOTE: combine() deletes fChild1 and/or fChild2 if they are not
+		// included in tree
 	if( !combine( newOp, val, tree, childOp1, val1, fChild1, 
 						childOp2, val2, fChild2 ) ) {
 		tree = NULL;
@@ -593,6 +635,8 @@ combine( OpKind &op, Value &val, ExprTree *&tree,
 						true, true, false, val );
 		if( val.IsBooleanValue() ) {
 			tree = NULL;
+			delete tree1;
+			delete tree2;
 			op = __NO_OP__;
 			return true;
 		}
@@ -616,7 +660,11 @@ combine( OpKind &op, Value &val, ExprTree *&tree,
 		return true;
 	} else if( ( tree1 && op1 == __NO_OP__ ) && ( tree2 && op2 == __NO_OP__ ) ){
 		// left and rightsons are trees only
-		if( !( newOp = MakeOperation( op, tree1, tree2 )) ) return false;
+		if( !( newOp = MakeOperation( op, tree1, tree2 )) ) {
+			delete tree1;
+			delete tree2;
+			return false;
+		}
 		tree = newOp;
 		op   = __NO_OP__;
 		return true;
@@ -664,7 +712,11 @@ combine( OpKind &op, Value &val, ExprTree *&tree,
 	if( op == op1 && op == op2 ) {
 		// same operators on both children . since op!=NO_OP, neither are op1, 
 		// op2.  so they both make tree and value contributions
-		if( !( newOp = MakeOperation( op, tree1, tree2 ) ) ) return false;
+		if( !( newOp = MakeOperation( op, tree1, tree2 ) ) ) {
+			delete tree1;
+			delete tree2;
+			return false;
+		}
 		_doOperation( op, val1, val2, dummy, true, true, false, val );
 		tree = newOp;
 		return true;
@@ -679,6 +731,8 @@ combine( OpKind &op, Value &val, ExprTree *&tree,
 			// rightson makes a tree contribution
 			Operation *local_newOp = MakeOperation( op, tree1, tree2 );
 			if( !local_newOp ) {
+				delete tree1;
+				delete tree2;
 				tree = NULL; op = __NO_OP__;	
 				return false;
 			}
@@ -697,6 +751,8 @@ combine( OpKind &op, Value &val, ExprTree *&tree,
 			// leftson makes a tree contribution
 			Operation *local_newOp = MakeOperation( op, tree1, tree2 );
 			if( !local_newOp ) {
+				delete tree1;
+				delete tree2;
 				tree = NULL; op = __NO_OP__;	
 				return false;
 			}
@@ -918,9 +974,18 @@ doArithmetic (OpKind op, Value &v1, Value &v2, Value &result)
 int Operation::
 doLogical (OpKind op, Value &v1, Value &v2, Value &result)
 {
+	bool		b1, b2;
+
+		// first coerece inputs to boolean if they are considered equivalent
+	if( !v1.IsBooleanValue( b1 ) && v1.IsBooleanValueEquiv( b1 ) ) {
+		v1.SetBooleanValue( b1 );
+	}
+	if( !v2.IsBooleanValue( b2 ) && v2.IsBooleanValueEquiv( b2 ) ) {
+		v2.SetBooleanValue( b2 );
+	}
+
 	Value::ValueType	vt1 = v1.GetType();
 	Value::ValueType	vt2 = v2.GetType();
-	bool		b1, b2;
 
 	if( vt1!=Value::UNDEFINED_VALUE && vt1!=Value::ERROR_VALUE && 
 			vt1!=Value::BOOLEAN_VALUE ) {
@@ -933,9 +998,6 @@ doLogical (OpKind op, Value &v1, Value &v2, Value &result)
 		return SIG_CHLD2;
 	}
 	
-	v1.IsBooleanValue( b1 );
-	v2.IsBooleanValue( b2 );
-
 	// handle unary operator
 	if (op == LOGICAL_NOT_OP) {
 		if( vt1 == Value::BOOLEAN_VALUE ) {
